﻿using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace APIClientLibrary
{
	/// <summary>
	/// The APIClient class wraps the complexity of interacting with the Sage API's
	/// It exposes four methods (GET, PUT, POST and DELETE) that all accept and 
	/// return strings.
	/// </summary>
    public class APIClient
	{
		/// <summary>
		/// Base URL for requests
		/// </summary>
		public string BaseUrl { get; set; }

		/// <summary>
		/// Your developer subscription key
		/// </summary>
		public string SubscriptionKey { get; set; }
		
		/// <summary>
		/// Sage ID Access Token
		/// </summary>
		public string AccessToken { get; set; }

		/// <summary>
		/// Site ID
		/// </summary>
		public string SiteID { get; set; }

		/// <summary>
		/// Company ID
		/// </summary>
		public string CompanyID { get; set; }

		/// <summary>
		/// Application Token (Murano Only)
		/// </summary>
		public string ApplicationToken { get; set; }

		/// <summary>
		/// Your developer signing key
		/// </summary>
		/// <remarks>No longer required. Kept only for backwards compatibility.</remarks>
		public string SigningKey { get; set; }

		/// <summary>
		/// Gets an entity
		/// </summary>
		/// <param name="url">The version and entity required</param>
		/// <example>
		/// v1/sites
		/// v1/departments?$orderby=code
		/// </example>
		/// <returns>string containing json representing entity requested</returns>
		public string Get(string url)
		{
			return GetAsync(url).Result;
		}

		/// <summary>
		/// Gets an entity
		/// </summary>
		/// <param name="url">The version and entity required</param>
		/// <example>
		/// v1/sites
		/// v1/departments?$orderby=code
		/// </example>
		/// <returns>string containing json representing entity requested</returns>
		public async Task<string> GetAsync(string url)
		{
			// Add the signing message handler to the HTTP Client
			// This deals with signing all requests made.
			using (HttpClient client = new HttpClient(new SigningMessageHandler(new HttpClientHandler())))
			{
				// Configure the HTTP client with the default headers
				SetupHttpClient(client);

				// Make the Get request
				HttpResponseMessage response = await client.GetAsync(url).ConfigureAwait(false);

				// Check the request was sucessful, if not an exception is thrown
				await CheckValidResponse(response);

				// Return the result of the request
				return await response.Content.ReadAsStringAsync().ConfigureAwait(false);
			}
		}

		/// <summary>
		/// Updates an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <param name="body">The json to put</param>
		/// <example>
		/// url : v1/departments/123
		/// body : json string
		/// </example>
		/// <returns>string containing the updated entity</returns>
		public string Put(string url, string body)
		{
			return PutAsync(url, body).Result;
		}

		/// <summary>
		/// Updates an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <param name="body">The json to put</param>
		/// <example>
		/// url : v1/departments/123
		/// body : json string
		/// </example>
		/// <returns>string containing the updated entity</returns>
		public async Task<string> PutAsync(string url, string body)
		{
			// Add the signing message handler to the HTTP Client
			// This deals with signing all requests made.
			using (HttpClient client = new HttpClient(new SigningMessageHandler(new HttpClientHandler())))
			{
				// Configure the HTTP client with the default headers
				SetupHttpClient(client);

				// Encode the message body
				StringContent content = new StringContent(body, Encoding.UTF8, "application/json");

				// Make the Put request
				HttpResponseMessage response = await client.PutAsync(url, content).ConfigureAwait(false);

				// Check the request was sucessful, if not an exception is thrown
				await CheckValidResponse(response);

				// Return the result of the request
				return await response.Content.ReadAsStringAsync().ConfigureAwait(false);
			}
		}

		/// <summary>
		/// Creates an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <param name="body">The json to post</param>
		/// <example>
		/// url : v1/departments
		/// body : json string
		/// </example>
		/// <returns>string containing the updated entity</returns>
		public string Post(string url, string body)
		{
			return PostAsync(url, body).Result;
		}

		/// <summary>
		/// Creates an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <param name="body">The json to post</param>
		/// <example>
		/// url : v1/departments
		/// body : json string
		/// </example>
		/// <returns>string containing the updated entity</returns>
		public async Task<string> PostAsync(string url, string body)
		{
            // Add the signing message handler to the HTTP Client
            // This deals with signing all requests made.
            using (HttpClient client = new HttpClient(new SigningMessageHandler(new HttpClientHandler())))
			{
				// Configure the HTTP client with the default headers
				SetupHttpClient(client);

				// Encode the message body
				StringContent content = new StringContent(body, Encoding.UTF8, "application/json");

                // Make the Post request
                HttpResponseMessage response = await client.PostAsync(url, content).ConfigureAwait(false);

                // Check the request was sucessful, if not an exception is thrown
                await CheckValidResponse(response);

				// Return the result of the request
				return await response.Content.ReadAsStringAsync().ConfigureAwait(false);
			}
		}

		/// <summary>
		/// Deletes an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <example>
		/// url : v1/departments/123
		/// </example>
		/// <returns>empty string, or string containing an error message</returns>
		public string Delete(string url)
		{
			return DeleteAsync(url).Result;
		}

		/// <summary>
		/// Deletes an entity
		/// </summary>
		/// <param name="url">The version and entity</param>
		/// <example>
		/// url : v1/departments/123
		/// </example>
		/// <returns>empty string, or string containing an error message</returns>
		public async Task<string> DeleteAsync(string url)
		{
			// Add the signing message handler to the HTTP Client
			// This deals with signing all requests made.
			using (HttpClient client = new HttpClient(new SigningMessageHandler(new HttpClientHandler())))
			{
				// Configure the HTTP client with the default header
				SetupHttpClient(client);

				// Make the Delete request
				HttpResponseMessage response = await client.DeleteAsync(url).ConfigureAwait(false);

				// Check the request was sucessful, if not an exception is thrown
				await CheckValidResponse(response);

				// Return the result of the request
				return await response.Content.ReadAsStringAsync().ConfigureAwait(false);
			}
		}

		/// <summary>
		/// Configures the HTTPClient for the base address and adds required headers
		/// </summary>
		/// <param name="client">The HttpClient</param>
		private void SetupHttpClient(HttpClient client)
		{
			// Add the Authorization header
			// value should be in format "Bearer" + single space + Token from SageID
			client.DefaultRequestHeaders.Add("Authorization", "Bearer " + AccessToken);
			
			// Configure the base URL for all requests
			client.BaseAddress = new Uri(BaseUrl);

			// Add the developer subscription key to the ocp-apim-subscription-key header
			client.DefaultRequestHeaders.Add("ocp-apim-subscription-key", SubscriptionKey);

			// Add the site id to the X-Site header
			client.DefaultRequestHeaders.Add("X-Site", SiteID);

			// Add the company id to the X-Company header
			client.DefaultRequestHeaders.Add("X-Company", CompanyID);

			#region Murano Specific Headers		
			// Murano specific headers - Will change when we have common headers

			// Add the site id to the tennant_id header
			client.DefaultRequestHeaders.Add("tenant_id", SiteID);

			// Add the Company id to the company_id header
			client.DefaultRequestHeaders.Add("company_id", CompanyID);
			
			if (string.IsNullOrEmpty(ApplicationToken) == false)
			{
				// if the Murano application token has been provided then add the
				// application token as the authorisation-sls-applicationClient-token header
				client.DefaultRequestHeaders.Add("authorisation-sls-applicationClient-token", ApplicationToken);
			}

			#endregion
		}

        /// <summary>
        /// Check if the HttpResponseMessage IsSuccessStatusCode. If not it reads the content error message and throws
        /// a APIClientException.
        /// </summary>
        /// <param name="response"></param>
        /// <returns></returns>
		private async Task CheckValidResponse(HttpResponseMessage response)
		{
			// Check the request was sucessful, if not an exception is thrown
			if (response != null && response.IsSuccessStatusCode == false)
			{
                string errorContent = string.Empty;

                // Read the response content which should contain the error message
                if (response.Content != null)
                {
                    errorContent = await response.Content.ReadAsStringAsync().ConfigureAwait(false);

                    // Dispose of the content
                    response.Content.Dispose();
                }

                // If we have a content message then deserialize it into and error message
				if (string.IsNullOrEmpty(errorContent) == false)
				{
					try
					{
						CustomError jsonResponse = JsonConvert.DeserializeObject<CustomError>(errorContent);

						errorContent = jsonResponse.Message;
					}
					catch (Exception)
					{
						// may not be the correct format
						errorContent = string.Empty;
					}
				}

                if (string.IsNullOrEmpty(errorContent))
                {
                    errorContent = "An unknown error has occurred";
                }

                // Throw a new APIClientException
                throw new APIClientException(response.StatusCode, errorContent);
			}
		}
	}

    /// <summary>
    /// Exception that is thrown when we have an error
    /// </summary>
	public class APIClientException : Exception
	{
		public HttpStatusCode StatusCode { get; private set; }

		public APIClientException(HttpStatusCode statusCode, string content)
			: base(content)
		{
			StatusCode = statusCode;
		}
	}

    /// <summary>
    /// Class used to deserialize the any exception message
    /// </summary>
    public class CustomError
    {
        public string Message { get; set; }
    }
}
