﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;

namespace APIClientLibrary
{
	/// <summary>
	/// Signs requests
	/// </summary>
	public class SignatureGenerator
	{
		/// <summary>
		/// Generates the signature from all the parameters passed in
		/// </summary>
		/// <param name="httpMethod">The HTTP method</param>
		/// <param name="url">The URL</param>
		/// <param name="body">The request message body (this will be null on a GET or DELETE)</param>
		/// <param name="secretKey">Your developer signing key</param>
		/// <param name="token">The authorization token</param>
		/// <param name="nonce">The nonce</param>
		/// <returns>The signature as a string</returns>
		public static string GenerateSignature(string httpMethod, Uri url, string body, string secretKey, string token, string nonce)
		{
			string baseString = GetBaseString(httpMethod, url, body, nonce);
			string signingKey = Uri.EscapeDataString(secretKey) + "&" + Uri.EscapeDataString(token);


			// Generate a hash based message authentication code based on the 
			// signing key and the entire message that has been parameterised and escaped
			string sig = GenerateHmac(signingKey, baseString);

			return sig;
		}

		internal static string GetBaseString(string httpMethod, Uri url, string body, string nonce)
		{
			// Uppercase the http method e.g. GET
			httpMethod = httpMethod.ToUpper();

			// Create an encoded string that contains the concatenated values of the Method, URL and request body
			string encodedParams = NormalizeParams(httpMethod, url, body);

			// Make the URL lowercase and Escape the string
			string encodedUri = Uri.EscapeDataString(url.GetLeftPart(UriPartial.Path).ToLower()); // this needs to be ToLower() to match the request check at the server

			// Escape the nonce
			string encodedNonce = Uri.EscapeDataString(nonce);

			// Build the signature base string to be signed with the Consumer Secret (Developer Key)
			string baseString = String.Format("{0}&{1}&{2}&{3}", httpMethod, encodedUri, encodedParams, encodedNonce);

			return baseString;
		}

		/// <summary>
		/// Generates a hash based message authentication code based on the signing key and 
		/// the entire message that has been parameterised and escaped
		/// </summary>
		/// <param name="signingKey">The developer signing key</param>
		/// <param name="baseString">The entire message that has been parameterised and escaped</param>
		/// <returns>The signature representing the message</returns>
		private static string GenerateHmac(string signingKey, string baseString)
		{
			HMACSHA1 hasher = new HMACSHA1(new ASCIIEncoding().GetBytes(signingKey));

			return Convert.ToBase64String(
				hasher.ComputeHash(
				new ASCIIEncoding().GetBytes(baseString)));
		}

        /// <summary>
        /// Creates a string that represents the HTTP method, the URL and the request body
        /// </summary>
        /// <param name="httpMethod">The HTTP method</param>
        /// <param name="url">The URL</param>
        /// <param name="requestBody">The request body</param>
        /// <returns>Am escaped and alphabetically sorted string representing the parameters</returns>
		private static string NormalizeParams(string httpMethod, Uri url, string body)
        {
			// Create a List of type KeyValuePair to contain the parameters
            IEnumerable<KeyValuePair<string, string>> kvpParams = new List<KeyValuePair<string, string>>();

			if (!string.IsNullOrWhiteSpace(url.Query))
            {
				// Extact all the query parameters from the url
                IEnumerable<KeyValuePair<string, string>> queryParams =
                  from p in url.Query.Substring(1).Split('&').AsEnumerable()
                  let key = Uri.EscapeDataString(p.Substring(0, p.IndexOf("=")))
                  let value = Uri.EscapeDataString(p.Substring(p.IndexOf("=") + 1))
                  select new KeyValuePair<string, string>(key, value);

				// Add each parameter and value found to the List of paramters
                kvpParams = kvpParams.Union(queryParams);
            }

            if (httpMethod == "POST" || httpMethod == "PUT")
            {
				var bytes = System.Text.Encoding.UTF8.GetBytes(body);
				body = System.Convert.ToBase64String(bytes);
				body = Uri.EscapeDataString(body);

				// Add the body to the List of parameters
                kvpParams = kvpParams.Union(new[]
                {
                    new KeyValuePair<string, string>("body", body)
                });
            }

            // Sort the parameters
            IEnumerable<string> sortedParams =
              from p in kvpParams
              orderby p.Key ascending, p.Value ascending
              select p.Key + "=" + p.Value;

            // Add the ampersand delimiter and then URL-encode
			string encodedParams = Uri.EscapeDataString(String.Join("&", sortedParams));
			
            return encodedParams;
        }
	}
}
