﻿using System;
using System.IO;
using System.Net.Http;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace APIClientLibrary
{
    /// <summary>
    /// Signing message handler is called when ever a http request is made to the API
	/// This signs the entire message: request type, url, paramters and body and adds 
	/// the signature as a header to the request.
    /// </summary>
	public class SigningMessageHandler : DelegatingHandler
    {
		/// <summary>
		/// The developers signing key
		/// </summary>
        /// <summary>
        /// Constructor to set up Signing Message hander.
		/// Passes in developer signing key and base handler.
        /// </summary>
        /// <param name="innerHandler">The base HTTPMessage handler</param>
        /// <param name="signingKey">The developers signing key</param>
		public SigningMessageHandler(HttpMessageHandler innerHandler) : base(innerHandler)
        {
        }

        /// <summary>
        /// Adds the x-nonce (number once) header - Used to stop replay of requests
		/// Generates a signature based on the request
		/// Adds the x-signature header to the message
		/// The API checks all requests for a valid signature, if a valid signature
		/// is not present then the request will fail as unauthorised.
        /// </summary>
        /// <param name="request">The HTTP request</param>
        /// <param name="cancellationToken">The cancellation token</param>
		/// <returns>Asynchronous Task of type HttpResponseMessage</returns>
		protected override Task<HttpResponseMessage> SendAsync(HttpRequestMessage request, CancellationToken cancellationToken)
        {
			// Adds the required unique header x-nonce
            string nonce = Guid.NewGuid().ToString("N");
            request.Headers.Add("X-Nonce", nonce);

            // Get the Sage ID authorization header from the request
			string authHeader = request.Headers.Authorization.ToString();
			// Strip the authorization header prefix 'Bearer ' from the token
            string token = authHeader.Replace("Bearer ", string.Empty);

			string requestContent = null;

			// If the request has content then read the content
			if (request.Content != null)
			{
				requestContent = request.Content.ReadAsStringAsync().Result;
			}

			// Call the GenerateSignature method passing in the following (which are all used to generate the signature)
			// The HTTP method e.g. GET, PUT, POST, DELETE
			// The request URI
			// The body of the request (this will be null for GET AND DELETE)
			// The nonce

			// Process the request
            return base.SendAsync(request, cancellationToken);
        }

    }

}
